import json
from httpretty import HTTPretty

from social_core.tests.backends.oauth import OAuth2Test
from social_core.tests.backends.open_id_connect import OpenIdConnectTestMixin

JWK_KEY = {
    'kty': 'RSA',
    'd': 'ZmswNokEvBcxW_Kvcy8mWUQOQCBdGbnM0xR7nhvGHC-Q24z3XAQWlMWbsmGc_R1o' \
         '_F3zK7DBlc3BokdRaO1KJirNmnHCw5TlnBlJrXiWpFBtVglUg98-4sRRO0VWnGXK' \
         'JPOkBQ6b_DYRO3b0o8CSpWowpiV6HB71cjXTqKPZf-aXU9WjCCAtxVjfIxgQFu5I' \
         '-G1Qah8mZeY8HK_y99L4f0siZcbUoaIcfeWBhxi14ODyuSAHt0sNEkhiIVBZE7QZ' \
         'm-SEP1ryT9VAaljbwHHPmg7NC26vtLZhvaBGbTTJnEH0ZubbN2PMzsfeNyoCIHy4' \
         '4QDSpQDCHfgcGOlHY_t5gQ',
    'e': 'AQAB',
    'use': 'sig',
    'kid': 'testkey',
    'alg': 'RS256',
    'n': 'pUfcJ8WFrVue98Ygzb6KEQXHBzi8HavCu8VENB2As943--bHPcQ-nScXnrRFAUg8' \
         'H5ZltuOcHWvsGw_AQifSLmOCSWJAPkdNb0w0QzY7Re8NrPjCsP58Tytp5LicF0Ao' \
         'Ag28UK3JioY9hXHGvdZsWR1Rp3I-Z3nRBP6HyO18pEgcZ91c9aAzsqu80An9X4DA' \
         'b1lExtZorvcd5yTBzZgr-MUeytVRni2lDNEpa6OFuopHXmg27Hn3oWAaQlbymd4g' \
         'ifc01oahcwl3ze2tMK6gJxa_TdCf1y99Yq6oilmVvZJ8kwWWnbPE-oDmOVPVnEyT' \
         'vYVCvN4rBT1DQ-x0F1mo2Q',
}

JWK_PUBLIC_KEY = {key: value for key, value in JWK_KEY.items() if key != 'd'}


class OktaOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.okta.OktaOAuth2'
    user_data_url = 'https://dev-000000.oktapreview.com/oauth2/v1/userinfo'
    expected_username = 'foo'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        'family_name': 'Bar',
        'sub': '101010101010101010101',
        'locale': 'en',
        'email_verified': True,
        'given_name': 'Foo',
        'email': 'foo@bar.com',
        'name': 'Foo Bar',
        'nickname': 'foobar',
        'middle_name': '',
        'profile': 'https://example.com/foo.bar',
        'zoneinfo': "America/Los_Angeles",
        "updated_at": 1311280970,
        "preferred_username": "foo",
    })

    def test_login(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_OKTA_OAUTH2_API_URL': 'https://dev-000000.oktapreview.com/oauth2'
        })
        self.do_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_OKTA_OAUTH2_API_URL': 'https://dev-000000.oktapreview.com/oauth2'
        })
        self.do_partial_pipeline()


class OktaOpenIdConnectTest(OpenIdConnectTestMixin, OAuth2Test):
    backend_path = \
        'social_core.backends.okta_openidconnect.OktaOpenIdConnect'
    user_data_url = \
        'https://dev-000000.oktapreview.com/oauth2/v1/userinfo'
    issuer = 'https://dev-000000.oktapreview.com/oauth2'
    openid_config_body = json.dumps({
        'issuer': 'https://dev-000000.oktapreview.com/oauth2',
        'authorization_endpoint': 'https://dev-000000.oktapreview.com/oauth2/v1/authorize',
        'token_endpoint': 'https://dev-000000.oktapreview.com/oauth2/v1/token',
        'userinfo_endpoint': 'https://dev-000000.oktapreview.com/oauth2/v1/userinfo',
        'jwks_uri': 'https://dev-000000.oktapreview.com/oauth2/v1/keys',
        'response_types_supported': [
            'code',
            'token',
            'id_token',
            'code token',
            'code id_token',
            'token id_token',
            'code token id_token',
            'none',
        ],
        'subject_types_supported': [
            'public',
        ],
        'id_token_signing_alg_values_supported': [
            'RS256',
        ],
        'scopes_supported': [
            'openid',
            'email',
            'profile',
        ],
        'token_endpoint_auth_methods_supported': [
            'client_secret_post',
            'client_secret_basic',
        ],
        'claims_supported': [
            'aud',
            'email',
            'email_verified',
            'exp',
            'family_name',
            'given_name',
            'iat',
            'iss',
            'locale',
            'name',
            'picture',
            'sub',
        ],
    })
    expected_username = 'foo'


    def setUp(self):
        super(OpenIdConnectTestMixin, self).setUp()
        # Settings for Okta
        self.strategy.set_settings({
            'SOCIAL_AUTH_OKTA_OPENIDCONNECT_API_URL': 'https://dev-000000.oktapreview.com/oauth2'
        })
        self.backend.OIDC_ENDPOINT = 'https://dev-000000.oktapreview.com/oauth2'

        self.key = JWK_KEY.copy()
        self.public_key = JWK_PUBLIC_KEY.copy()

        HTTPretty.register_uri(HTTPretty.GET,
                               self.backend.OIDC_ENDPOINT + '/.well-known/openid-configuration',
                               status=200,
                               body=self.openid_config_body)
        oidc_config = json.loads(self.openid_config_body)

        def jwks(_request, _uri, headers):
            return 200, headers, json.dumps({'keys': [self.key]})

        HTTPretty.register_uri(HTTPretty.GET,
                               oidc_config.get('jwks_uri'),
                               status=200,
                               body=json.dumps({'keys': [self.public_key]}))

        self.backend.JWKS_URI = oidc_config.get('jwks_uri')
        self.backend.ID_TOKEN_ISSUER = oidc_config.get('issuer')



"""
Copyright (c) 2017 Noderabbit Inc., d.b.a. Appsembler

All rights reserved.

MIT License

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import json
import jwt

from time import time
from httpretty import HTTPretty
from jwt.algorithms import RSAAlgorithm

from .oauth import OAuth2Test


# Dummy private and private keys:
RSA_PUBLIC_JWT_KEY = {
    # https://github.com/jpadilla/pyjwt/blob/06f461a/tests/keys/jwk_rsa_pub.json
    'kty': 'RSA',
    'kid': 'bilbo.baggins@hobbiton.example',
    'use': 'sig',
    'n': 'n4EPtAOCc9AlkeQHPzHStgAbgs7bTZLwUBZdR8_KuKPEHLd4rHVTeT-O-XV2jRojdNhxJWTDvNd7nqQ0VEiZQHz_AJmSCpMaJMRBSFKrKb2wqVwGU_NsYOYL-QtiWN2lbzcEe6XC0dApr5ydQLrHqkHHig3RBordaZ6Aj-oBHqFEHYpPe7Tpe-OfVfHd1E6cS6M1FZcD1NNLYD5lFHpPI9bTwJlsde3uhGqC0ZCuEHg8lhzwOHrtIQbS0FVbb9k3-tVTU4fg_3L_vniUFAKwuCLqKnS2BYwdq_mzSnbLY7h_qixoR7jig3__kRhuaxwUkRz5iaiQkqgc5gHdrNP5zw',
    'e': 'AQAB'
}

RSA_PRIVATE_JWT_KEY = {
    # https://github.com/jpadilla/pyjwt/blob/06f461a/tests/keys/jwk_rsa_key.json
    'kty': 'RSA',
    'kid': 'bilbo.baggins@hobbiton.example',
    'use': 'sig',
    'n': 'n4EPtAOCc9AlkeQHPzHStgAbgs7bTZLwUBZdR8_KuKPEHLd4rHVTeT-O-XV2jRojdNhxJWTDvNd7nqQ0VEiZQHz_AJmSCpMaJMRBSFKrKb2wqVwGU_NsYOYL-QtiWN2lbzcEe6XC0dApr5ydQLrHqkHHig3RBordaZ6Aj-oBHqFEHYpPe7Tpe-OfVfHd1E6cS6M1FZcD1NNLYD5lFHpPI9bTwJlsde3uhGqC0ZCuEHg8lhzwOHrtIQbS0FVbb9k3-tVTU4fg_3L_vniUFAKwuCLqKnS2BYwdq_mzSnbLY7h_qixoR7jig3__kRhuaxwUkRz5iaiQkqgc5gHdrNP5zw',
    'e': 'AQAB',
    'd': 'bWUC9B-EFRIo8kpGfh0ZuyGPvMNKvYWNtB_ikiH9k20eT-O1q_I78eiZkpXxXQ0UTEs2LsNRS-8uJbvQ-A1irkwMSMkK1J3XTGgdrhCku9gRldY7sNA_AKZGh-Q661_42rINLRCe8W-nZ34ui_qOfkLnK9QWDDqpaIsA-bMwWWSDFu2MUBYwkHTMEzLYGqOe04noqeq1hExBTHBOBdkMXiuFhUq1BU6l-DqEiWxqg82sXt2h-LMnT3046AOYJoRioz75tSUQfGCshWTBnP5uDjd18kKhyv07lhfSJdrPdM5Plyl21hsFf4L_mHCuoFau7gdsPfHPxxjVOcOpBrQzwQ',
    'p': '3Slxg_DwTXJcb6095RoXygQCAZ5RnAvZlno1yhHtnUex_fp7AZ_9nRaO7HX_-SFfGQeutao2TDjDAWU4Vupk8rw9JR0AzZ0N2fvuIAmr_WCsmGpeNqQnev1T7IyEsnh8UMt-n5CafhkikzhEsrmndH6LxOrvRJlsPp6Zv8bUq0k',
    'q': 'uKE2dh-cTf6ERF4k4e_jy78GfPYUIaUyoSSJuBzp3Cubk3OCqs6grT8bR_cu0Dm1MZwWmtdqDyI95HrUeq3MP15vMMON8lHTeZu2lmKvwqW7anV5UzhM1iZ7z4yMkuUwFWoBvyY898EXvRD-hdqRxHlSqAZ192zB3pVFJ0s7pFc',
    'dp': 'B8PVvXkvJrj2L-GYQ7v3y9r6Kw5g9SahXBwsWUzp19TVlgI-YV85q1NIb1rxQtD-IsXXR3-TanevuRPRt5OBOdiMGQp8pbt26gljYfKU_E9xn-RULHz0-ed9E9gXLKD4VGngpz-PfQ_q29pk5xWHoJp009Qf1HvChixRX59ehik',
    'dq': 'CLDmDGduhylc9o7r84rEUVn7pzQ6PF83Y-iBZx5NT-TpnOZKF1pErAMVeKzFEl41DlHHqqBLSM0W1sOFbwTxYWZDm6sI6og5iTbwQGIC3gnJKbi_7k_vJgGHwHxgPaX2PnvP-zyEkDERuf-ry4c_Z11Cq9AqC2yeL6kdKT1cYF8',
    'qi': '3PiqvXQN0zwMeE-sBvZgi289XP9XCQF3VWqPzMKnIgQp7_Tugo6-NZBKCQsMf3HaEGBjTVJs_jcK8-TRXvaKe-7ZMaQj8VfBdYkssbu0NKDDhjJ-GtiseaDVWt7dcH0cfwxgFUHpQh7FoCrjFJ6h6ZEpMF6xmujs4qMpPz8aaI4'
}


class AzureADOAuth2Test(OAuth2Test):
    AUTH_KEY = 'abcdef12-1234-9876-0000-abcdef098765'
    EXPIRES_IN = 3600
    AUTH_TIME = int(time())
    EXPIRES_ON = AUTH_TIME + EXPIRES_IN

    backend_path = 'social_core.backends.azuread_b2c.AzureADB2COAuth2'
    expected_username = 'FooBar'
    refresh_token_body = json.dumps({
        'access_token': 'foobar-new-token',
        'token_type': 'bearer',
        'expires_in': EXPIRES_IN,
        'refresh_token': 'foobar-new-refresh-token',
        'scope': 'identity'
    })

    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer',
        'id_token': jwt.encode(
            key=RSAAlgorithm.from_jwk(json.dumps(RSA_PRIVATE_JWT_KEY)),
            headers={
                'kid': RSA_PRIVATE_JWT_KEY['kid'],
            },
            algorithm='RS256',
            payload={
                'aud': AUTH_KEY,
                'auth_time': AUTH_TIME,
                'country': 'Axphain',
                'emails': [
                    'foobar@example.com'
                ],
                'exp': EXPIRES_ON,
                'family_name': 'Bar',
                'given_name': 'Foo',
                'iat': AUTH_TIME,
                'iss': 'https://login.microsoftonline.com/9a9a9a9a-1111-5555-0000-bc24adfdae00/v2.0/',
                'name': 'FooBar',
                'nbf': AUTH_TIME,
                'oid': '11223344-5566-7788-9999-aabbccddeeff',
                'postalCode': '00000',
                'sub': '11223344-5566-7788-9999-aabbccddeeff',
                'tfp': 'B2C_1_SignIn',
                'ver': '1.0',
        }).decode('ascii'),
        'expires_in': EXPIRES_IN,
        'expires_on': EXPIRES_ON,
        'not_before': AUTH_TIME,
    })

    def extra_settings(self):
        settings = super(AzureADOAuth2Test, self).extra_settings()
        settings.update({
            'SOCIAL_AUTH_' + self.name + '_POLICY': 'b2c_1_signin',
            'SOCIAL_AUTH_' + self.name + '_KEY': self.AUTH_KEY,
            'SOCIAL_AUTH_' + self.name + '_TENANT_ID': 'footenant.onmicrosoft.com',
        })
        return settings

    def setUp(self):
        super(AzureADOAuth2Test, self).setUp()

        keys_url = 'https://login.microsoftonline.com/footenant.onmicrosoft.com/discovery/v2.0/keys?p=b2c_1_signin'
        keys_body = json.dumps({
            'keys': [{
                # Dummy public key that pairs with `access_token_body` key:
                # https://github.com/jpadilla/pyjwt/blob/06f461a/tests/keys/jwk_rsa_pub.json
                'kty': 'RSA',
                'kid': 'bilbo.baggins@hobbiton.example',
                'use': 'sig',
                'n': 'n4EPtAOCc9AlkeQHPzHStgAbgs7bTZLwUBZdR8_KuKPEHLd4rHVTeT-O-X'
                     'V2jRojdNhxJWTDvNd7nqQ0VEiZQHz_AJmSCpMaJMRBSFKrKb2wqVwGU_Ns'
                     'YOYL-QtiWN2lbzcEe6XC0dApr5ydQLrHqkHHig3RBordaZ6Aj-oBHqFEHY'
                     'pPe7Tpe-OfVfHd1E6cS6M1FZcD1NNLYD5lFHpPI9bTwJlsde3uhGqC0ZCu'
                     'EHg8lhzwOHrtIQbS0FVbb9k3-tVTU4fg_3L_vniUFAKwuCLqKnS2BYwdq_'
                     'mzSnbLY7h_qixoR7jig3__kRhuaxwUkRz5iaiQkqgc5gHdrNP5zw',
                'e': 'AQAB',
        }],
        })
        HTTPretty.register_uri(HTTPretty.GET, keys_url, status=200, body=keys_body)

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()

    def test_refresh_token(self):
        user, social = self.do_refresh_token()
        self.assertEqual(social.extra_data['access_token'], 'foobar-new-token')



"""
MediaWiki OAuth1 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/mediawiki.html
"""

import re
import time
import six
import requests
import jwt

from six import b
from six.moves.urllib.parse import parse_qs, urlencode, urlparse
from requests_oauthlib import OAuth1

from .oauth import BaseOAuth1
from ..exceptions import AuthException


def force_unicode(value):
    """
    Return string in unicode.
    """
    if isinstance(value, six.text_type):
        return value
    else:
        if six.PY3:
            return str(value, "unicode-escape")
        else:
            return unicode(value, "unicode-escape")


class MediaWiki(BaseOAuth1):
    """
    Handles the handshake with Mediawiki and fetching of user data.
    """
    name = 'mediawiki'
    MEDIAWIKI_URL = 'https://meta.wikimedia.org/w/index.php'
    SOCIAL_AUTH_MEDIAWIKI_CALLBACK = 'oob'
    LEEWAY = 10.0

    def unauthorized_token(self):
        """
        Return request for unauthorized token (first stage)

        Mediawiki request token is requested from e.g.:
         * https://en.wikipedia.org/w/index.php?title=Special:OAuth/initiate
        """
        params = self.request_token_extra_arguments()
        params.update(self.get_scope_argument())
        params['title'] = 'Special:OAuth/initiate'
        key, secret = self.get_key_and_secret()
        decoding = None if six.PY3 else 'utf-8'
        response = self.request(
            self.setting('MEDIAWIKI_URL'),
            params=params,
            auth=OAuth1(
                key,
                secret,
                callback_uri=self.setting('CALLBACK'),
                decoding=decoding
            ),
            method=self.REQUEST_TOKEN_METHOD
        )

        if response.content.decode().startswith('Error'):
            raise AuthException(self, response.content.decode())

        return response.content.decode()

    def oauth_authorization_request(self, token):
        """
        Generates the URL for the authorization link
        """
        if not isinstance(token, dict):
            token = parse_qs(token)

        oauth_token = token.get(self.OAUTH_TOKEN_PARAMETER_NAME)[0]
        state = self.get_or_create_state()
        base_url = self.setting('MEDIAWIKI_URL')

        return '{0}?{1}'.format(base_url, urlencode({
            'title': 'Special:Oauth/authenticate',
            self.OAUTH_TOKEN_PARAMETER_NAME: oauth_token,
            self.REDIRECT_URI_PARAMETER_NAME: self.get_redirect_uri(state)
        }))

    def access_token(self, token):
        """
        Fetches the Mediawiki access token.
        """
        auth_token = self.oauth_auth(token)

        response = requests.post(
            url=self.setting('MEDIAWIKI_URL'),
            params={'title': 'Special:Oauth/token'},
            auth=auth_token
        )
        credentials = parse_qs(response.content)
        oauth_token_key = credentials.get(b('oauth_token'))[0]
        oauth_token_secret = credentials.get(b('oauth_token_secret'))[0]
        oauth_token_key = oauth_token_key.decode()
        oauth_token_secret = oauth_token_secret.decode()

        return {
            'oauth_token': oauth_token_key,
            'oauth_token_secret': oauth_token_secret
        }

    def get_user_details(self, response):
        """
        Gets the user details from Special:OAuth/identify
        """
        key, secret = self.get_key_and_secret()
        access_token = response['access_token']

        auth = OAuth1(key, client_secret=secret,
                      resource_owner_key=access_token['oauth_token'],
                      resource_owner_secret=access_token['oauth_token_secret'])

        req_resp = requests.post(url=self.setting('MEDIAWIKI_URL'),
                                 params={'title': 'Special:OAuth/identify'},
                                 auth=auth)

        try:
            identity = jwt.decode(req_resp.content, secret,
                                  audience=key, algorithms=['HS256'],
                                  leeway=self.LEEWAY)
        except jwt.InvalidTokenError as exception:
            raise AuthException(
                self,
                'An error occurred while trying to read json ' +
                'content: {0}'.format(exception)
            )

        issuer = urlparse(identity['iss']).netloc
        expected_domain = urlparse(self.setting('MEDIAWIKI_URL')).netloc

        if not issuer == expected_domain:
            raise AuthException(
                self,
                'Unexpected issuer {0}, expected {1}'.format(
                    issuer,
                    expected_domain
                )
            )

        now = time.time()
        issued_at = float(identity['iat'])
        if not now >= (issued_at - self.LEEWAY):
            raise AuthException(
                self,
                'Identity issued {0} seconds in the future'.format(
                    issued_at - now
                )
            )

        authorization_header = force_unicode(
            req_resp.request.headers['Authorization']
        )
        request_nonce = re.search(r'oauth_nonce="(.*?)"',
                                  authorization_header).group(1)

        if identity['nonce'] != request_nonce:
            raise AuthException(
                self,
                'Replay attack detected: {0} != {1}'.format(
                    identity['nonce'],
                    request_nonce
                )
            )

        return {
            'username': identity['username'],
            'userID': identity['sub'],
            'email': identity.get('email'),
            'confirmed_email': identity.get('confirmed_email'),
            'editcount': identity.get('editcount'),
            'rights': identity.get('rights'),
            'groups': identity.get('groups'),
            'registered': identity.get('registered'),
            'blocked': identity.get('blocked')
        }

    def get_user_id(self, details, response):
        """
        Get the unique Mediawiki user ID.
        """
        return details['userID']



"""
Mozilla Persona authentication backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/persona.html
"""
from ..utils import handle_http_errors
from .base import BaseAuth
from ..exceptions import AuthFailed, AuthMissingParameter


class PersonaAuth(BaseAuth):
    """BrowserID authentication backend"""
    name = 'persona'

    def get_user_id(self, details, response):
        """Use BrowserID email as ID"""
        return details['email']

    def get_user_details(self, response):
        """Return user details, BrowserID only provides Email."""
        # {'status': 'okay',
        #  'audience': 'localhost:8000',
        #  'expires': 1328983575529,
        #  'email': 'name@server.com',
        #  'issuer': 'browserid.org'}
        email = response['email']
        return {'username': email.split('@', 1)[0],
                'email': email,
                'fullname': '',
                'first_name': '',
                'last_name': ''}

    def extra_data(self, user, uid, response, details=None, *args, **kwargs):
        """Return users extra data"""
        return {'audience': response['audience'],
                'issuer': response['issuer']}

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        if 'assertion' not in self.data:
            raise AuthMissingParameter(self, 'assertion')

        response = self.get_json('https://browserid.org/verify', data={
            'assertion': self.data['assertion'],
            'audience': self.strategy.request_host()
        }, method='POST')
        if response.get('status') == 'failure':
            raise AuthFailed(self)
        kwargs.update({'response': response, 'backend': self})
        return self.strategy.authenticate(*args, **kwargs)



import json

from httpretty import HTTPretty

from .oauth import OAuth2Test


class AtlassianOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.atlassian.AtlassianOAuth2'
    tenant_url = 'https://api.atlassian.com/oauth/token/accessible-resources'
    user_data_url = 'https://api.atlassian.com/ex/jira/FAKED_CLOUD_ID/rest/api/2/myself'
    expected_username = 'erlich'
    access_token_body = json.dumps({
        'access_token': 'aviato',
        'token_type': 'bearer'
    })
    tenant_data_body = json.dumps([
      {
          "id": "FAKED_CLOUD_ID",
          "name": "bachmanity.com",
          "avatarUrl": "https://bachmanity.atlassian.net/avatars/240/site.png",
          "scopes": [
            "read:jira-user"
          ]
      }
    ])
    user_data_body = json.dumps({
        "self": "http://bachmanity.atlassian.net/rest/api/3/user?username=erlich",
        "key": "erlich",
        "accountId": "99:27935d01-92a7-4687-8272-a9b8d3b2ae2e",
        "name": "erlich",
        "emailAddress": "erlich@bachmanity.com",
        "avatarUrls": {
            "48x48": "http://bachmanity.atlassian.net/secure/useravatar?size=large&ownerId=erlich",
            "24x24": "http://bachmanity.atlassian.net/secure/useravatar?size=small&ownerId=erlich",
            "16x16": "http://bachmanity.atlassian.net/secure/useravatar?size=xsmall&ownerId=erlich",
            "32x32": "http://bachmanity.atlassian.net/secure/useravatar?size=medium&ownerId=erlich"
        },
        "displayName": "Erlich Bachman",
        "active": True,
        "timeZone": "Australia/Sydney",
        "groups": {
            "size": 3,
            "items": []
        },
        "applicationRoles": {
            "size": 1,
            "items": []
        }
    })

    def auth_handlers(self, start_url):
        target_url = super(AtlassianOAuth2Test, self).auth_handlers(start_url)
        HTTPretty.register_uri(HTTPretty.GET,
                               self.tenant_url,
                               body=self.tenant_data_body,
                               content_type='application/json')
        return target_url

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



from .oauth import BaseOAuth2


# This provides a backend for python-social-auth. This should not be confused
# with officially battle.net offerings. This piece of code is not officially
# affiliated with Blizzard Entertainment, copyrights to their respective
# owners. See http://us.battle.net/en/forum/topic/13979588015 for more details.


class BattleNetOAuth2(BaseOAuth2):
    """ battle.net Oauth2 backend"""
    name = 'battlenet-oauth2'
    ID_KEY = 'accountId'
    REDIRECT_STATE = False
    AUTHORIZATION_URL = 'https://eu.battle.net/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://eu.battle.net/oauth/token'
    ACCESS_TOKEN_METHOD = 'POST'
    REVOKE_TOKEN_METHOD = 'GET'
    DEFAULT_SCOPE = ['wow.profile']
    EXTRA_DATA = [
        ('refresh_token', 'refresh_token', True),
        ('expires_in', 'expires'),
        ('token_type', 'token_type', True)
    ]

    def get_characters(self, access_token):
        """
        Fetches the character list from the battle.net API. Returns list of
        characters or empty list if the request fails.
        """
        params = {'access_token': access_token}
        if self.setting('API_LOCALE'):
            params['locale'] = self.setting('API_LOCALE')

        response = self.get_json(
            'https://eu.api.battle.net/wow/user/characters',
            params=params
        )
        return response.get('characters') or []

    def get_user_details(self, response):
        """ Return user details from Battle.net account """
        return {'battletag': response.get('battletag')}

    def user_data(self, access_token, *args, **kwargs):
        """ Loads user data from service """
        return self.get_json(
            'https://eu.api.battle.net/account/user',
            params={'access_token': access_token}
        )



from .oauth import BaseOAuth2


class MailChimpOAuth2(BaseOAuth2):
    """MailChimp OAuth2 authentication backend"""
    name = 'mailchimp'
    AUTHORIZATION_URL = 'https://login.mailchimp.com/oauth2/authorize'
    ACCESS_TOKEN_URL = 'https://login.mailchimp.com/oauth2/token'
    METADATA_URL = 'https://login.mailchimp.com/oauth2/metadata'
    ACCESS_TOKEN_METHOD = 'POST'
    STATE_PARAMETER = False
    REDIRECT_STATE = False
    ID_KEY = 'user_id'
    EXTRA_DATA = [
        ('accountname', 'accountname'),
        ('api_endpoint', 'api_endpoint'),
        ('role', 'role'),
        ('login', 'login')
    ]

    def get_user_details(self, response):
        """Return user details from a Mailchimp metadata response"""
        return {
            'username': response['login']['login_name'],
            'email': response['login']['email']
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data and datacenter information from service"""
        return self.get_json(self.METADATA_URL, headers={
          'Authorization': 'OAuth ' + access_token
        })



from .oauth import BaseOAuth2


class EventbriteOAuth2(BaseOAuth2):
    """Eventbrite OAuth2 authentication backend"""
    name = 'eventbrite'
    AUTHORIZATION_URL = 'https://www.eventbrite.com/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://www.eventbrite.com/oauth/token'
    METADATA_URL = 'https://www.eventbriteapi.com/v3/users/me'
    ACCESS_TOKEN_METHOD = 'POST'
    STATE_PARAMETER = False
    REDIRECT_STATE = False

    def get_user_details(self, response):
        """Return user details from an Eventbrite metadata response"""
        email = next(iter(filter(lambda x: x['primary'], response['emails'])))['email']

        return {
            'username': email,
            'email': email,
            'first_name': response['first_name'],
            'last_name': response['last_name']
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data and datacenter information from service"""
        return self.get_json(self.METADATA_URL, headers={
          'Authorization': 'Bearer ' + access_token
        })



# coding:utf-8
# author:hepochen@gmail.com  https://github.com/hepochen
"""
Weibo OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/weibo.html
"""
from .oauth import BaseOAuth2


class WeiboOAuth2(BaseOAuth2):
    """Weibo (of sina) OAuth authentication backend"""
    name = 'weibo'
    ID_KEY = 'uid'
    AUTHORIZATION_URL = 'https://api.weibo.com/oauth2/authorize'
    REQUEST_TOKEN_URL = 'https://api.weibo.com/oauth2/request_token'
    ACCESS_TOKEN_URL = 'https://api.weibo.com/oauth2/access_token'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ('id', 'id'),
        ('name', 'username'),
        ('profile_image_url', 'profile_image_url'),
        ('gender', 'gender')
    ]

    def get_user_details(self, response):
        """Return user details from Weibo. API URL is:
        https://api.weibo.com/2/users/show.json/?uid=<UID>&access_token=<TOKEN>
        """
        if self.setting('DOMAIN_AS_USERNAME'):
            username = response.get('domain', '')
        else:
            username = response.get('name', '')
        fullname, first_name, last_name = self.get_user_names(
            first_name=response.get('screen_name', '')
        )
        return {'username': username,
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def get_uid(self, access_token):
        """Return uid by access_token"""
        data = self.get_json(
            'https://api.weibo.com/oauth2/get_token_info',
            method='POST',
            params={'access_token': access_token}
        )
        return data['uid']

    def user_data(self, access_token, response=None, *args, **kwargs):
        """Return user data"""
        # If user id was not retrieved in the response, then get it directly
        # from weibo get_token_info endpoint
        uid = response and response.get('uid') or self.get_uid(access_token)
        user_data = self.get_json(
            'https://api.weibo.com/2/users/show.json',
            params={'access_token': access_token, 'uid': uid}
        )
        user_data['uid'] = uid
        return user_data



import json

from .oauth import OAuth2Test


class QuizletOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.quizlet.QuizletOAuth2'
    expected_username = 'foo_bar'

    access_token_body = json.dumps({
        "access_token": "EE1IDxytP04tJ767GbjH7ED9PpGmYvL",
        "token_type": "Bearer",
        "expires_in": 3600,
        "scope": "read",
        'user_id': 'foo_bar'
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json

from .oauth import OAuth2Test


class SimpleLoginOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.simplelogin.SimpleLoginOAuth2'
    user_data_url = 'https://app.simplelogin.io/oauth2/userinfo'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        "client": "Continental",
        "email": "john@wick.com",
        "email_verified": True,
        "id": 1,
        "name": "John Wick",
        "avatar_url": "http://wick.com/john.png"
    })
    expected_username = "john@wick.com"

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import base64
import json

from cryptography.x509 import load_pem_x509_certificate
from cryptography.hazmat.backends import default_backend
from jwt import DecodeError, ExpiredSignature, decode as jwt_decode

from ..exceptions import AuthTokenError
from .azuread import AzureADOAuth2

"""
Copyright (c) 2015 Microsoft Open Technologies, Inc.

All rights reserved.

MIT License

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

"""
Azure AD OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/azuread.html

See https://nicksnettravels.builttoroam.com/post/2017/01/24/Verifying-Azure-Active-Directory-JWT-Tokens.aspx
for verifying JWT tokens.
"""


class AzureADTenantOAuth2(AzureADOAuth2):
    name = 'azuread-tenant-oauth2'
    OPENID_CONFIGURATION_URL = \
        'https://login.microsoftonline.com/{tenant_id}/.well-known/openid-configuration'
    AUTHORIZATION_URL = \
        'https://login.microsoftonline.com/{tenant_id}/oauth2/authorize'
    ACCESS_TOKEN_URL = 'https://login.microsoftonline.com/{tenant_id}/oauth2/token'
    JWKS_URL = 'https://login.microsoftonline.com/{tenant_id}/discovery/keys'

    @property
    def tenant_id(self):
        return self.setting('TENANT_ID', 'common')

    def openid_configuration_url(self):
        return self.OPENID_CONFIGURATION_URL.format(tenant_id=self.tenant_id)

    def authorization_url(self):
        return self.AUTHORIZATION_URL.format(tenant_id=self.tenant_id)

    def access_token_url(self):
        return self.ACCESS_TOKEN_URL.format(tenant_id=self.tenant_id)

    def jwks_url(self):
        return self.JWKS_URL.format(tenant_id=self.tenant_id)

    def get_certificate(self, kid):
        # retrieve keys from jwks_url
        resp = self.request(self.jwks_url(), method='GET')
        resp.raise_for_status()

        # find the proper key for the kid
        for key in resp.json()['keys']:
            if key['kid'] == kid:
                x5c = key['x5c'][0]
                break
        else:
            raise DecodeError('Cannot find kid={}'.format(kid))

        certificate = '-----BEGIN CERTIFICATE-----\n' \
                      '{}\n' \
                      '-----END CERTIFICATE-----'.format(x5c)

        return load_pem_x509_certificate(certificate.encode(),
                                         default_backend())

    def get_user_id(self, details, response):
        """Use subject (sub) claim as unique id."""
        return response.get('sub')

    def user_data(self, access_token, *args, **kwargs):
        response = kwargs.get('response')
        id_token = response.get('id_token')

        # decode the JWT header as JSON dict
        jwt_header = json.loads(
            base64.b64decode(id_token.split('.', 1)[0]).decode()
        )

        # get key id and algorithm
        key_id = jwt_header['kid']
        algorithm = jwt_header['alg']

        try:
            # retrieve certificate for key_id
            certificate = self.get_certificate(key_id)

            return jwt_decode(
                id_token,
                key=certificate.public_key(),
                algorithms=algorithm,
                audience=self.setting('KEY')
            )
        except (DecodeError, ExpiredSignature) as error:
            raise AuthTokenError(self, error)



"""
Steam OpenId backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/steam.html
"""
from .open_id import OpenIdAuth
from ..exceptions import AuthFailed


USER_INFO = 'http://api.steampowered.com/ISteamUser/GetPlayerSummaries/v0002/?'


class SteamOpenId(OpenIdAuth):
    name = 'steam'
    URL = 'https://steamcommunity.com/openid'

    def get_user_id(self, details, response):
        """Return user unique id provided by service"""
        return self._user_id(response)

    def get_user_details(self, response):
        player = self.get_json(USER_INFO, params={
            'key': self.setting('API_KEY'),
            'steamids': self._user_id(response)
        })
        if len(player['response']['players']) > 0:
            player = player['response']['players'][0]
            details = {'username': player.get('personaname'),
                       'email': '',
                       'fullname': '',
                       'first_name': '',
                       'last_name': '',
                       'player': player}
        else:
            details = {}
        return details

    def consumer(self):
        # Steam seems to support stateless mode only, ignore store
        if not hasattr(self, '_consumer'):
            self._consumer = self.create_consumer()
        return self._consumer

    def _user_id(self, response):
        user_id = response.identity_url.rsplit('/', 1)[-1]
        if not user_id.isdigit():
            raise AuthFailed(self, 'Missing Steam Id')
        return user_id



"""
SurveyMonkey OAuth2 backend, docs at:
    https://developer.surveymonkey.com/api/v3/#authentication
"""
from .oauth import BaseOAuth2


class SurveyMonkeyOAuth2(BaseOAuth2):
    """SurveyMonkey OAuth2 authentication backend"""
    name = 'surveymonkey'
    AUTHORIZATION_URL = 'https://api.surveymonkey.com/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://api.surveymonkey.com/oauth/token'
    ACCESS_TOKEN_METHOD = 'POST'
    USER_DATA_URL = '/v3/users/me'
    STATE_PARAMETER = False
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ('access_url', 'access_url'),
    ]

    def get_user_details(self, response):
        """Return user details from a SurveyMonkey /users/me response"""
        response["name"] = response['first_name'] + ' ' + response['last_name']
        return response

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data information from service"""
        base_url = kwargs["response"]["access_url"]
        return self.get_json(base_url + self.USER_DATA_URL, headers={
          'Authorization': 'bearer ' + access_token
        })



import json

from .oauth import OAuth2Test


class KakaoOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.kakao.KakaoOAuth2'
    user_data_url = 'https://kapi.kakao.com/v2/user/me'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar'
    })
    user_data_body = json.dumps({
        'id': '101010101',
        'properties': {
            'nickname': 'foobar',
            'thumbnail_image': 'http://mud-kage.kakao.co.kr/14/dn/btqbh1AKmRf/'
                               'ujlHpQhxtMSbhKrBisrxe1/o.jpg',
            'profile_image': 'http://mud-kage.kakao.co.kr/14/dn/btqbjCnl06Q/'
                             'wbMJSVAUZB7lzSImgGdsoK/o.jpg'
        },
        'kakao_account': {
            'email': 'a'
        }
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Auth0 implementation based on:
https://auth0.com/docs/quickstart/webapp/django/01-login
"""
from jose import jwt

from .oauth import BaseOAuth2


class Auth0OAuth2(BaseOAuth2):
    """Auth0 OAuth authentication backend"""
    name = 'auth0'
    SCOPE_SEPARATOR = ' '
    ACCESS_TOKEN_METHOD = 'POST'
    EXTRA_DATA = [
        ('picture', 'picture')
    ]

    def api_path(self, path=''):
        """Build API path for Auth0 domain"""
        return 'https://{domain}/{path}'.format(domain=self.setting('DOMAIN'),
                                                path=path)

    def authorization_url(self):
        return self.api_path('authorize')

    def access_token_url(self):
        return self.api_path('oauth/token')

    def get_user_id(self, details, response):
        """Return current user id."""
        return details['user_id']

    def get_user_details(self, response):
        # Obtain JWT and the keys to validate the signature
        id_token = response.get('id_token')
        jwks = self.get_json(self.api_path('.well-known/jwks.json'))
        issuer = self.api_path()
        audience = self.setting('KEY')  # CLIENT_ID
        payload = jwt.decode(id_token,
                             jwks,
                             algorithms=['RS256'],
                             audience=audience,
                             issuer=issuer)
        fullname, first_name, last_name = self.get_user_names(payload['name'])
        return {'username': payload['nickname'],
                'email': payload['email'],
                'email_verified': payload.get('email_verified', False),
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name,
                'picture': payload['picture'],
                'user_id': payload['sub']}



import json

from .oauth import OAuth2Test


class PhabricatorOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.phabricator.PhabricatorOAuth2'
    user_data_url = 'https://secure.phabricator.com/api/user.whoami'
    expected_username = 'user'
    access_token_body = json.dumps({
        'access_token': 'loremipsumdolorsitametenim',
        'token_type': 'Bearer',
        'expires_in': 7200,
        'refresh_token': 'foobar',
    })

    user_data_body = json.dumps({
        'phid': 'PHID-USER-jbfcsj7c6nkt2tv3trb6',
        'userName': 'user',
        'realName': 'FirstName LastName',
        'image': 'https://secure.phabricator.com/file/data/4qjdpmvca4wwkfw2wevc'
                 '/PHID-FILE-t37vxezr54fjuvbrblkp/alphanumeric_lato-white_U.png'
                 '-_3f674d-255%2C255%2C255%2C0.7.png',
        'uri': 'https://secure.phabricator.com/p/user/',
        'roles': ['admin', 'verified', 'approved', 'activated'],
        'primaryEmail': 'user@example.com',
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()


class PhabricatorCustomDomainOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.phabricator.PhabricatorOAuth2'
    user_data_url = 'https://example.com/api/user.whoami'
    expected_username = 'user'
    access_token_body = json.dumps({
        'access_token': 'loremipsumdolorsitametenim',
        'token_type': 'Bearer',
        'expires_in': 7200,
        'refresh_token': 'foobar',
    })

    user_data_body = json.dumps({
        'phid': 'PHID-USER-jbfcsj7c6nkt2tv3trb6',
        'userName': 'user',
        'realName': 'FirstName LastName',
        'image': 'https://example.com/file/data/4qjdpmvca4wwkfw2wevc/PHID-FILE-'
                 't37vxezr54fjuvbrblkp/alphanumeric_lato-white_U.png-_3f674d-25'
                 '5%2C255%2C255%2C0.7.png',
        'uri': 'https://example.com/p/user/',
        'roles': ['admin', 'verified', 'approved', 'activated'],
        'primaryEmail': 'user@example.com',
    })

    def test_login(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_PHABRICATOR_API_URL': 'https://example.com',
        })
        self.do_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings({
            'SOCIAL_AUTH_PHABRICATOR_API_URL': 'https://example.com',
        })
        self.do_partial_pipeline()



"""
Okta OAuth2 and OpenIdConnect:
    https://python-social-auth.readthedocs.io/en/latest/backends/okta.html
"""
from .okta import OktaOAuth2
from .open_id_connect import OpenIdConnectAuth


class OktaOpenIdConnect(OktaOAuth2, OpenIdConnectAuth):
    """Okta OpenID-Connect authentication backend"""
    name = 'okta-openidconnect'
    REDIRECT_STATE = False
    ACCESS_TOKEN_METHOD = 'POST'
    RESPONSE_TYPE = 'code'



DEFAULT_AUTH_PIPELINE = (
    # Get the information we can about the user and return it in a simple
    # format to create the user instance later. On some cases the details are
    # already part of the auth response from the provider, but sometimes this
    # could hit a provider API.
    'social_core.pipeline.social_auth.social_details',

    # Get the social uid from whichever service we're authing thru. The uid is
    # the unique identifier of the given user in the provider.
    'social_core.pipeline.social_auth.social_uid',

    # Verifies that the current auth process is valid within the current
    # project, this is where emails and domains whitelists are applied (if
    # defined).
    'social_core.pipeline.social_auth.auth_allowed',

    # Checks if the current social-account is already associated in the site.
    'social_core.pipeline.social_auth.social_user',

    # Make up a username for this person, appends a random string at the end if
    # there's any collision.
    'social_core.pipeline.user.get_username',

    # Send a validation email to the user to verify its email address.
    # 'social_core.pipeline.mail.mail_validation',

    # Associates the current social details with another user account with
    # a similar email address.
    # 'social_core.pipeline.social_auth.associate_by_email',

    # Create a user account if we haven't found one yet.
    'social_core.pipeline.user.create_user',

    # Create the record that associated the social account with this user.
    'social_core.pipeline.social_auth.associate_user',

    # Populate the extra_data field in the social record with the values
    # specified by settings (and the default ones like access_token, etc).
    'social_core.pipeline.social_auth.load_extra_data',

    # Update the user record with any changed info from the auth service.
    'social_core.pipeline.user.user_details'
)

DEFAULT_DISCONNECT_PIPELINE = (
    # Verifies that the social association can be disconnected from the current
    # user (ensure that the user login mechanism is not compromised by this
    # disconnection).
    'social_core.pipeline.disconnect.allowed_to_disconnect',

    # Collects the social associations to disconnect.
    'social_core.pipeline.disconnect.get_entries',

    # Revoke any access_token when possible.
    'social_core.pipeline.disconnect.revoke_tokens',

    # Removes the social associations.
    'social_core.pipeline.disconnect.disconnect'
)



import time

from jwt import DecodeError, ExpiredSignature, decode as jwt_decode

from ..exceptions import AuthTokenError
from .oauth import BaseOAuth2

"""
Copyright (c) 2015 Microsoft Open Technologies, Inc.

All rights reserved.

MIT License

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

"""
Azure AD OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/azuread.html
"""


class AzureADOAuth2(BaseOAuth2):
    name = 'azuread-oauth2'
    SCOPE_SEPARATOR = ' '
    AUTHORIZATION_URL = \
        'https://login.microsoftonline.com/common/oauth2/authorize'
    ACCESS_TOKEN_URL = 'https://login.microsoftonline.com/common/oauth2/token'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False
    DEFAULT_SCOPE = ['openid', 'profile', 'user_impersonation']
    EXTRA_DATA = [
        ('access_token', 'access_token'),
        ('id_token', 'id_token'),
        ('refresh_token', 'refresh_token'),
        ('expires_in', 'expires'),
        ('expires_on', 'expires_on'),
        ('not_before', 'not_before'),
        ('given_name', 'first_name'),
        ('family_name', 'last_name'),
        ('token_type', 'token_type')
    ]

    def get_user_id(self, details, response):
        """Use upn as unique id"""
        return response.get('upn')

    def get_user_details(self, response):
        """Return user details from Azure AD account"""
        fullname, first_name, last_name = (
            response.get('name', ''),
            response.get('given_name', ''),
            response.get('family_name', '')
        )
        return {'username': fullname,
                'email': response.get('upn'),
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        response = kwargs.get('response')
        if response and response.get('id_token'):
            id_token = response.get('id_token')
        else:
            id_token = access_token
            
        try:
            decoded_id_token = jwt_decode(id_token, verify=False)
        except (DecodeError, ExpiredSignature) as de:
            raise AuthTokenError(self, de)
        return decoded_id_token

    def auth_extra_arguments(self):
        """Return extra arguments needed on auth process. The defaults can be
        overridden by GET parameters."""
        extra_arguments = super(AzureADOAuth2, self).auth_extra_arguments()
        resource = self.setting('RESOURCE')
        if resource:
            extra_arguments.update({'resource': resource})
        return extra_arguments

    def extra_data(self, user, uid, response, details=None, *args, **kwargs):
        """Return access_token and extra defined names to store in
        extra_data field"""
        data = super(AzureADOAuth2, self).extra_data(user, uid, response,
            details, *args, **kwargs)
        data['resource'] = self.setting('RESOURCE')
        return data

    def refresh_token_params(self, token, *args, **kwargs):
        return {
            'client_id': self.setting('KEY'),
            'client_secret': self.setting('SECRET'),
            'refresh_token': token,
            'grant_type': 'refresh_token',
            'resource': self.setting('RESOURCE')
        }

    def get_auth_token(self, user_id):
        """Return the access token for the given user, after ensuring that it
        has not expired, or refreshing it if so."""
        user = self.get_user(user_id=user_id)
        access_token = user.social_user.access_token
        expires_on = user.social_user.extra_data['expires_on']
        if expires_on <= int(time.time()):
            new_token_response = self.refresh_token(token=access_token)
            access_token = new_token_response['access_token']
        return access_token



from ..exceptions import AuthAlreadyAssociated, AuthException, AuthForbidden


def social_details(backend, details, response, *args, **kwargs):
    return {'details': dict(backend.get_user_details(response), **details)}


def social_uid(backend, details, response, *args, **kwargs):
    return {'uid': backend.get_user_id(details, response)}


def auth_allowed(backend, details, response, *args, **kwargs):
    if not backend.auth_allowed(response, details):
        raise AuthForbidden(backend)


def social_user(backend, uid, user=None, *args, **kwargs):
    provider = backend.name
    social = backend.strategy.storage.user.get_social_auth(provider, uid)
    if social:
        if user and social.user != user:
            msg = 'This account is already in use.'
            raise AuthAlreadyAssociated(backend, msg)
        elif not user:
            user = social.user
    return {'social': social,
            'user': user,
            'is_new': user is None,
            'new_association': social is None}


def associate_user(backend, uid, user=None, social=None, *args, **kwargs):
    if user and not social:
        try:
            social = backend.strategy.storage.user.create_social_auth(
                user, uid, backend.name
            )
        except Exception as err:
            if not backend.strategy.storage.is_integrity_error(err):
                raise
            # Protect for possible race condition, those bastard with FTL
            # clicking capabilities, check issue #131:
            #   https://github.com/omab/django-social-auth/issues/131
            return social_user(backend, uid, user, *args, **kwargs)
        else:
            return {'social': social,
                    'user': social.user,
                    'new_association': True}


def associate_by_email(backend, details, user=None, *args, **kwargs):
    """
    Associate current auth with a user with the same email address in the DB.

    This pipeline entry is not 100% secure unless you know that the providers
    enabled enforce email verification on their side, otherwise a user can
    attempt to take over another user account by using the same (not validated)
    email address on some provider.  This pipeline entry is disabled by
    default.
    """
    if user:
        return None

    email = details.get('email')
    if email:
        # Try to associate accounts registered with the same email address,
        # only if it's a single object. AuthException is raised if multiple
        # objects are returned.
        users = list(backend.strategy.storage.user.get_users_by_email(email))
        if len(users) == 0:
            return None
        elif len(users) > 1:
            raise AuthException(
                backend,
                'The given email address is associated with another account'
            )
        else:
            return {'user': users[0],
                    'is_new': False}


def load_extra_data(backend, details, response, uid, user, *args, **kwargs):
    social = kwargs.get('social') or \
             backend.strategy.storage.user.get_social_auth(backend.name, uid)
    if social:
        extra_data = backend.extra_data(user, uid, response, details,
                                        *args, **kwargs)
        social.set_extra_data(extra_data)



"""
Shimmering Oauth
"""
from .oauth import BaseOAuth2


class ShimmeringOAuth2(BaseOAuth2):
    """Shimmering Verify OAuth2 authentication backend"""
    name = 'shimmering'
    ID_KEY = 'id'
    AUTHORIZATION_URL = 'http://developers.shimmeringverify.com/o/authorize/'
    ACCESS_TOKEN_URL = 'http://developers.shimmeringverify.com/o/token/'
    ACCESS_TOKEN_METHOD = 'POST'

    def get_user_details(self, response):
        """Return user details from Shimmering"""
        first_name = response.get('first_name')
        last_name = response.get('last_name')
        email = response.get('email')
        username = response.get('username')
        fullname = '{} {}'.format(first_name, last_name)
        return {
            'username': username,
            'fullname': fullname,
            'first_name': first_name,
            'last_name': last_name,
            'email': email,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        headers = {'Authorization': 'Bearer %s' % access_token}
        return self.get_json(
            'http://developers.shimmeringverify.com/user_info/',
            headers=headers
        )



from ..strategy import BaseStrategy, BaseTemplateStrategy


TEST_URI = 'http://myapp.com'
TEST_HOST = 'myapp.com'


class Redirect(object):
    def __init__(self, url):
        self.url = url


class TestTemplateStrategy(BaseTemplateStrategy):
    def render_template(self, tpl, context):
        return tpl

    def render_string(self, html, context):
        return html


class TestStrategy(BaseStrategy):
    DEFAULT_TEMPLATE_STRATEGY = TestTemplateStrategy

    def __init__(self, storage, tpl=None):
        self._request_data = {}
        self._settings = {}
        self._session = {}
        super(TestStrategy, self).__init__(storage, tpl)

    def redirect(self, url):
        return Redirect(url)

    def get_setting(self, name):
        """Return value for given setting name"""
        return self._settings[name]

    def html(self, content):
        """Return HTTP response with given content"""
        return content

    def render_html(self, tpl=None, html=None, context=None):
        """Render given template or raw html with given context"""
        return tpl or html

    def request_data(self, merge=True):
        """Return current request data (POST or GET)"""
        return self._request_data

    def request_host(self):
        """Return current host value"""
        return TEST_HOST

    def request_is_secure(self):
        """ Is the request using HTTPS? """
        return False

    def request_path(self):
        """ path of the current request """
        return ''

    def request_port(self):
        """ Port in use for this request """
        return 80

    def request_get(self):
        """ Request GET data """
        return self._request_data.copy()

    def request_post(self):
        """ Request POST data """
        return self._request_data.copy()

    def session_get(self, name, default=None):
        """Return session value for given key"""
        return self._session.get(name, default)

    def session_set(self, name, value):
        """Set session value for given key"""
        self._session[name] = value

    def session_pop(self, name):
        """Pop session value for given key"""
        return self._session.pop(name, None)

    def build_absolute_uri(self, path=None):
        """Build absolute URI with given (optional) path"""
        path = path or ''
        if path.startswith('http://') or path.startswith('https://'):
            return path
        return TEST_URI + path

    def set_settings(self, values):
        self._settings.update(values)

    def set_request_data(self, values, backend):
        self._request_data.update(values)
        backend.data = self._request_data

    def remove_from_request_data(self, name):
        self._request_data.pop(name, None)

    def authenticate(self, *args, **kwargs):
        user = super(TestStrategy, self).authenticate(*args, **kwargs)
        if isinstance(user, self.storage.user.user_model()):
            self.session_set('username', user.username)
        return user

    def get_pipeline(self, backend=None):
        return self.setting(
            'PIPELINE',
            (
                'social_core.pipeline.social_auth.social_details',
                'social_core.pipeline.social_auth.social_uid',
                'social_core.pipeline.social_auth.auth_allowed',
                'social_core.pipeline.social_auth.social_user',
                'social_core.pipeline.user.get_username',
                'social_core.pipeline.social_auth.associate_by_email',
                'social_core.pipeline.user.create_user',
                'social_core.pipeline.social_auth.associate_user',
                'social_core.pipeline.social_auth.load_extra_data',
                'social_core.pipeline.user.user_details'
            ),
            backend
        )



"""
ThisIsMyJam OAuth1 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/thisismyjam.html
"""
from .oauth import BaseOAuth1


class ThisIsMyJamOAuth1(BaseOAuth1):
    """ThisIsMyJam OAuth1 authentication backend"""
    name = 'thisismyjam'
    REQUEST_TOKEN_URL = 'http://www.thisismyjam.com/oauth/request_token'
    AUTHORIZATION_URL = 'http://www.thisismyjam.com/oauth/authorize'
    ACCESS_TOKEN_URL = 'http://www.thisismyjam.com/oauth/access_token'
    REDIRECT_URI_PARAMETER_NAME = 'oauth_callback'

    def get_user_details(self, response):
        """Return user details from ThisIsMyJam account"""
        info = response.get('person')
        fullname, first_name, last_name = self.get_user_names(
            info.get('fullname')
        )
        return {
            'username': info.get('name'),
            'email': '',
            'fullname': fullname,
            'first_name': first_name,
            'last_name': last_name
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json('http://api.thisismyjam.com/1/verify.json',
                             auth=self.oauth_auth(access_token))



"""
Discord Auth OAuth2 backend, docs at:
    https://discordapp.com/developers/docs/topics/oauth2
"""
from .oauth import BaseOAuth2


class DiscordOAuth2(BaseOAuth2):
    name = 'discord'
    AUTHORIZATION_URL = 'https://discordapp.com/api/oauth2/authorize'
    ACCESS_TOKEN_URL = 'https://discordapp.com/api/oauth2/token'
    ACCESS_TOKEN_METHOD = 'POST'
    REVOKE_TOKEN_URL = 'https://discordapp.com/api/oauth2/token/revoke'
    REVOKE_TOKEN_METHOD = 'GET'
    DEFAULT_SCOPE = ['identify']
    SCOPE_SEPARATOR = '+'
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ('expires_in', 'expires'),
        ('refresh_token', 'refresh_token')
    ]

    def get_user_details(self, response):
        return {'username': response.get('username'),
                'email': response.get('email') or ''}

    def user_data(self, access_token, *args, **kwargs):
        url = 'https://discordapp.com/api/users/@me'
        auth_header = {"Authorization": "Bearer %s" % access_token}
        return self.get_json(url, headers=auth_header)



"""
Uber OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/uber.html
"""
from .oauth import BaseOAuth2


class UberOAuth2(BaseOAuth2):
    name = 'uber'
    ID_KEY = 'uuid'
    SCOPE_SEPARATOR = ' '
    AUTHORIZATION_URL = 'https://login.uber.com/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://login.uber.com/oauth/token'
    ACCESS_TOKEN_METHOD = 'POST'

    def auth_complete_credentials(self):
        return self.get_key_and_secret()

    def get_user_details(self, response):
        """Return user details from Uber account"""
        email = response.get('email', '')
        fullname, first_name, last_name = self.get_user_names(
            '',
            response.get('first_name', ''),
            response.get('last_name', '')
        )
        return {'username': email,
                'email': email,
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        response = kwargs.pop('response')
        return self.get_json('https://api.uber.com/v1/me', headers={
            'Authorization': '{0} {1}'.format(response.get('token_type'),
                                              access_token)
        })



import json

from .oauth import OAuth2Test


class NaverOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.naver.NaverOAuth2'
    user_data_url = 'https://openapi.naver.com/v1/nid/me'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer',
    })

    user_data_content_type = 'text/json'
    user_data_body = json.dumps({
        "resultcode": "00",
        "message": "success",
        "response": {
            "email": "openapi@naver.com",
            "nickname": "foobar",
            "profile_image": "https://ssl.pstatic.net/static/pwe/address/nodata_33x33.gif",
            "age": "40-49",
            "gender": "F",
            "id": "32742776",
            "name": "foobar",
            "birthday": "10-01",
        }
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json

from .oauth import OAuth2Test


class StravaOAuthTest(OAuth2Test):
    backend_path = 'social_core.backends.strava.StravaOAuth'
    user_data_url = 'https://www.strava.com/api/v3/athlete'
    expected_username = 'marianne_v'
    access_token_body = json.dumps({
        "token_type": "Bearer",
        "expires_at": 1572805000,
        "expires_in": 227615,
        "refresh_token": "f51defab4632d27255dd0d106504dfd7568fd1df6",
        "access_token": "83ebeabdec09f6670863766f792ead24d61fe3f9",
        "athlete": {
            "id": 1234567890987654321,
            "username": "marianne_v",
            "resource_state": 2,
            "firstname": "Marianne",
            "lastname": "V.",
            "city": "Francisco",
            "state": "California",
            "country": "United States",
            "sex": "F",
            "premium": "true",
            "summit": "true",
            "created_at": "2017-11-14T02:30:05Z",
            "updated_at": "2018-02-06T19:32:20Z",
            "badge_type_id": 4,
            "profile_medium": "https://xxxxxx.cloudfront.net/pictures/athletes/123456789/123456789/2/medium.jpg",
            "profile": "https://xxxxx.cloudfront.net/pictures/athletes/123456789/123456789/2/large.jpg",
            "friend": "null",
            "follower": "null"
        }
    })
    user_data_body = json.dumps({
      "id": 1234567890987654321,
      "username": "marianne_v",
      "resource_state": 3,
      "firstname": "Marianne",
      "lastname": "V.",
      "city": "San Francisco",
      "state": "CA",
      "country": "US",
      "sex": "F",
      "premium": "true",
      "created_at": "2017-11-14T02:30:05Z",
      "updated_at": "2018-02-06T19:32:20Z",
      "badge_type_id": 4,
      "profile_medium": "https://xxxxxx.cloudfront.net/pictures/athletes/123456789/123456789/2/medium.jpg",
      "profile": "https://xxxxx.cloudfront.net/pictures/athletes/123456789/123456789/2/large.jpg",
      "friend": "null",
      "follower": "null",
      "follower_count": 5,
      "friend_count": 5,
      "mutual_friend_count": 0,
      "athlete_type": 1,
      "date_preference": "%m/%d/%Y",
      "measurement_preference": "feet",
      "clubs": [],
      "ftp": "null",
      "weight": 0,
      "bikes": [],
      "shoes": []
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json

from .oauth import OAuth2Test


class PodioOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.podio.PodioOAuth2'
    user_data_url = 'https://api.podio.com/user/status'
    expected_username = 'user_1010101010'
    access_token_body = json.dumps({
        'token_type': 'bearer',
        'access_token': '11309ea9016a4ad99f1a3bcb9bc7a9d1',
        'refresh_token': '52d01df8b9ac46a4a6be1333d9f81ef2',
        'expires_in': 28800,
        'ref': {
            'type': 'user',
            'id': 1010101010,
        }
    })
    user_data_body = json.dumps({
        'user': {
            'user_id': 1010101010,
            'activated_on': '2012-11-22 09:37:21',
            'created_on': '2012-11-21 12:23:47',
            'locale': 'en_GB',
            'timezone': 'Europe/Copenhagen',
            'mail': 'foo@bar.com',
            'mails': [
                {
                    'disabled': False,
                    'mail': 'foobar@example.com',
                    'primary': False,
                    'verified': True
                }, {
                    'disabled': False,
                    'mail': 'foo@bar.com',
                    'primary': True,
                    'verified': True
                }
            ],
            # more properties ...
        },
        'profile': {
            'last_seen_on': '2013-05-16 12:21:13',
            'link': 'https://podio.com/users/1010101010',
            'name': 'Foo Bar',
            # more properties ...
        }
        # more properties ...
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Beats backend, docs at:
    https://developer.beatsmusic.com/docs
"""
import base64

from ..utils import handle_http_errors
from .oauth import BaseOAuth2


class BeatsOAuth2(BaseOAuth2):
    name = 'beats'
    SCOPE_SEPARATOR = ' '
    ID_KEY = 'user_context'
    AUTHORIZATION_URL = \
        'https://partner.api.beatsmusic.com/v1/oauth2/authorize'
    ACCESS_TOKEN_URL = 'https://partner.api.beatsmusic.com/oauth2/token'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False

    def get_user_id(self, details, response):
        return response['result'][BeatsOAuth2.ID_KEY]

    def auth_headers(self):
        return {
            'Authorization': 'Basic {0}'.format(base64.urlsafe_b64encode(
                ('{0}:{1}'.format(*self.get_key_and_secret()).encode())
            ))
        }

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        self.process_error(self.data)
        response = self.request_access_token(
            self.ACCESS_TOKEN_URL,
            data=self.auth_complete_params(self.validate_state()),
            headers=self.auth_headers(),
            method=self.ACCESS_TOKEN_METHOD
        )
        self.process_error(response)
        # mashery wraps in jsonrpc
        if response.get('jsonrpc', None):
            response = response.get('result', None)
        return self.do_auth(response['access_token'], response=response,
                            *args, **kwargs)

    def get_user_details(self, response):
        """Return user details from Beats account"""
        response = response['result']
        fullname, first_name, last_name = self.get_user_names(
            response.get('display_name')
        )
        return {'username': response.get('id'),
                'email': response.get('email'),
                'fullname': fullname,
                'first_name': first_name,
                'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            'https://partner.api.beatsmusic.com/v1/api/me',
            headers={'Authorization': 'Bearer {0}'.format(access_token)}
        )



import json

from .oauth import OAuth2Test


class PatreonOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.patreon.PatreonOAuth2'
    user_data_url = 'https://www.patreon.com/api/oauth2/v2/identity?fields%5Buser%5D=about,created,email,first_name,full_name,image_url,last_name,social_connections,thumb_url,url,vanity'
    expected_username = 'JohnInterwebs'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer',
    })
    user_data_body = json.dumps({
        "data": {
            "relationships": {
                "pledges": {
                    "data": [{
                        "type": "pledge", "id": "123456"
                    }]
                }
            },
            "attributes": {
                "last_name": "Interwebs",
                "is_suspended": False,
                "has_password": True,
                "full_name": "John Interwebs",
                "is_nuked": False,
                "first_name": "John",
                "social_connections": {
                    "spotify": None,
                    "discord": None,
                    "twitter": None,
                    "youtube": None,
                    "facebook": None,
                    "deviantart": None,
                    "twitch": None
                },
                "twitter": None,
                "is_email_verified": True,
                "facebook_id": None,
                "email": "john@example.com",
                "facebook": None,
                "thumb_url": "https://c8.patreon.com/100/123456",
                "vanity": None,
                "about": None,
                "is_deleted": False,
                "created": "2017-05-05T05:16:34+00:00",
                "url": "https://www.patreon.com/user?u=123456",
                "gender": 0,
                "youtube": None,
                "discord_id": None,
                "image_url": "https://c8.patreon.com/400/123456",
                "twitch": None
            },
            "type": "user",
            "id": "123456"
        }
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Ubuntu One OpenId backend
"""
from .open_id import OpenIdAuth


class UbuntuOpenId(OpenIdAuth):
    name = 'ubuntu'
    URL = 'https://login.ubuntu.com'

    def get_user_id(self, details, response):
        """
        Return user unique id provided by service. For Ubuntu One
        the nickname should be original.
        """
        return details['nickname']



"""
Patreon OAuth2 backend
https://www.patreon.com/platform/documentation/oauth
"""
from .oauth import BaseOAuth2


class PatreonOAuth2(BaseOAuth2):
    """Patreon OAuth2 authentication backend"""
    name = 'patreon'
    AUTHORIZATION_URL = 'https://www.patreon.com/oauth2/authorize'
    ACCESS_TOKEN_URL = 'https://www.patreon.com/api/oauth2/token'
    REVOKE_TOKEN_URL = 'https://www.patreon.com/oauth2/revoke'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False
    ID_KEY = 'id'
    EXTRA_DATA = [
        ('id', 'id'),
    ]

    def get_user_details(self, response):
        details = response['attributes']
        return {
            'username': details.get('full_name'),
            'email': details.get('email'),
            'fullname': details.get('full_name'),
            'first_name': details.get('first_name'),
            'last_name': details.get('last_name'),
        }

    def user_data(self, access_token, *args, **kwargs):
        return self.get_api(access_token, 'identity')['data']

    def get_api(self, access_token, suffix):
        return self.get_json(
            'https://www.patreon.com/api/oauth2/v2/{}'.format(suffix),
            headers=self.get_auth_header(access_token)
        )

    def get_auth_header(self, access_token):
        return {'Authorization': 'Bearer {0}'.format(access_token)}



import json

from .oauth import OAuth2Test


class FlatOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.flat.FlatOAuth2'
    user_data_url = 'https://api.flat.io/v2/me'
    expected_username = 'vincent'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        "id": "541a137946fd04d57cb2e3c0",
        "username": "vincent",
        "name": "Vincent Foo",
        "printableName": "Vincent Foo",
        "bio": "Foo bio",
        "instruments": [],
        "picture": "https://flat-prod-public.s3.amazonaws.com/00000000/a0d2cb86-ab1e-4fdb-9286-dd94aa6d386c.jpeg",
        "registrationDate": "2014-09-17T23:04:25.042Z",
        "htmlUrl": "https://flat.io/vincent",
        "starredScoresCount": 85,
        "likedScoresCount": 85,
        "followersCount": 183,
        "followingCount": 52,
        "ownedPublicScoresCount": 15,
        "isPowerUser": True
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()


"""
Jawbone OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/jawbone.html
"""
from ..utils import handle_http_errors
from .oauth import BaseOAuth2
from ..exceptions import AuthCanceled, AuthUnknownError


class JawboneOAuth2(BaseOAuth2):
    name = 'jawbone'
    AUTHORIZATION_URL = 'https://jawbone.com/auth/oauth2/auth'
    ACCESS_TOKEN_URL = 'https://jawbone.com/auth/oauth2/token'
    SCOPE_SEPARATOR = ' '
    REDIRECT_STATE = False

    def get_user_id(self, details, response):
        return response['data']['xid']

    def get_user_details(self, response):
        """Return user details from Jawbone account"""
        data = response['data']
        fullname, first_name, last_name = self.get_user_names(
            first_name=data.get('first', ''),
            last_name=data.get('last', '')
        )
        return {
            'username': first_name + ' ' + last_name,
            'fullname': fullname,
            'first_name': first_name,
            'last_name': last_name,
            'dob': data.get('dob', ''),
            'gender': data.get('gender', ''),
            'height': data.get('height', ''),
            'weight': data.get('weight', '')
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            'https://jawbone.com/nudge/api/users/@me',
            headers={'Authorization': 'Bearer ' + access_token},
        )

    def process_error(self, data):
        error = data.get('error')
        if error:
            if error == 'access_denied':
                raise AuthCanceled(self)
            else:
                raise AuthUnknownError(self, 'Jawbone error was {0}'.format(
                    error
                ))
        return super(JawboneOAuth2, self).process_error(data)

    def auth_complete_params(self, state=None):
        client_id, client_secret = self.get_key_and_secret()
        return {
            'grant_type': 'authorization_code',  # request auth code
            'code': self.data.get('code', ''),  # server response code
            'client_id': client_id,
            'client_secret': client_secret,
        }

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        self.process_error(self.data)
        response = self.request_access_token(
            self.ACCESS_TOKEN_URL,
            params=self.auth_complete_params(self.validate_state()),
            headers=self.auth_headers(),
            method=self.ACCESS_TOKEN_METHOD
        )
        self.process_error(response)
        return self.do_auth(response['access_token'], response=response,
                            *args, **kwargs)



"""
Flat OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/flat.html
"""
from .oauth import BaseOAuth2


class FlatOAuth2(BaseOAuth2):
    """Flat OAuth2"""
    name = 'flat'
    DEFAULT_SCOPE = ['account.public_profile']
    AUTHORIZATION_URL = 'https://flat.io/auth/oauth'
    ACCESS_TOKEN_URL = 'https://api.flat.io/oauth/access_token'
    ACCESS_TOKEN_METHOD = 'POST'

    def get_user_id(self, details, response):
        return response.get('id')

    def get_user_details(self, response):
        """Return user details from Flat account"""
        return {
            'email': response.get('email'),
            'username': response.get('username'),
            'fullname': response.get('printableName')
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json('https://api.flat.io/v2/me', headers={
            'Authorization': 'Bearer ' + access_token
        })



import json

from .oauth import OAuth2Test


class SketchfabOAuth2Test(OAuth2Test):
    backend_path = 'social_core.backends.sketchfab.SketchfabOAuth2'
    user_data_url = 'https://sketchfab.com/v2/users/me'
    expected_username = 'foobar'
    access_token_body = json.dumps({
        'access_token': 'foobar',
        'token_type': 'bearer'
    })
    user_data_body = json.dumps({
        'uid': '42',
        'email': 'foo@bar.com',
        'displayName': 'foo bar',
        'username': 'foobar',
        'apiToken': 'XXX'
    })

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Created on May 13, 2014

@author: Yong Zhang (zyfyfe@gmail.com)
"""

import json

from ..utils import parse_qs
from .oauth import BaseOAuth2


class QQOAuth2(BaseOAuth2):
    name = 'qq'
    ID_KEY = 'openid'
    AUTHORIZE_URL = 'https://graph.qq.com/oauth2.0/authorize'
    ACCESS_TOKEN_URL = 'https://graph.qq.com/oauth2.0/token'
    AUTHORIZATION_URL = 'https://graph.qq.com/oauth2.0/authorize'
    OPENID_URL = 'https://graph.qq.com/oauth2.0/me'
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ('nickname', 'username'),
        ('figureurl_qq_1', 'profile_image_url'),
        ('gender', 'gender')
    ]

    def get_user_details(self, response):
        """
        Return user detail from QQ account sometimes nickname will duplicate
        with another qq account, to avoid this issue it's possible to use
        openid as username.
        """
        if self.setting('USE_OPENID_AS_USERNAME', False):
            username = response.get('openid', '')
        else:
            username = response.get('nickname', '')

        fullname, first_name, last_name = self.get_user_names(
            first_name=response.get('nickname', '')
        )

        return {
            'username': username,
            'fullname': fullname,
            'first_name': first_name,
            'last_name': last_name
        }

    def get_openid(self, access_token):
        response = self.request(self.OPENID_URL, params={
            'access_token': access_token
        })
        content = response.content.decode()
        data = json.loads(content[10:-3])
        return data['openid']

    def user_data(self, access_token, *args, **kwargs):
        openid = self.get_openid(access_token)
        response = self.get_json(
            'https://graph.qq.com/user/get_user_info', params={
                'access_token': access_token,
                'oauth_consumer_key': self.setting('KEY'),
                'openid': openid
            }
        )
        response['openid'] = openid
        return response

    def request_access_token(self, url, data, *args, **kwargs):
        response = self.request(url, *args, **kwargs)
        return parse_qs(response.content)



"""
SimpleLogin OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/simplelogin.html
"""

from .oauth import BaseOAuth2


class SimpleLoginOAuth2(BaseOAuth2):
    """SimpleLogin OAuth authentication backend"""
    name = 'simplelogin'
    AUTHORIZATION_URL = 'https://app.simplelogin.io/oauth2/authorize'
    ACCESS_TOKEN_URL = 'https://app.simplelogin.io/oauth2/token'
    ACCESS_TOKEN_METHOD = 'POST'
    REDIRECT_STATE = False
    STATE_PARAMETER = True
    SEND_USER_AGENT = True
    EXTRA_DATA = [
        ('name', 'name'),
        ('email', 'email'),
        ('avatar_url', 'avatar_url'),
    ]

    # endpoint to get user info
    USERINFO_URL = "https://app.simplelogin.io/oauth2/userinfo"

    def get_user_details(self, response):
        """Return user details from SimpleLogin account"""
        fullname, first_name, last_name = self.get_user_names(
            response.get('name')
        )
        return {
            'username': response.get('email'),
            'email': response.get('email'),
            'fullname': fullname,
            'first_name': first_name,
            'last_name': last_name}

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(self.USERINFO_URL, params={
            'access_token': access_token
        })



"""
HubSpot OAuth2 backend, docs at:
    https://developers.hubspot.com/docs/methods/oauth2/oauth2-overview
"""
import time
import datetime
from .oauth import BaseOAuth2


class HubSpotOAuth2(BaseOAuth2):
    """HubSpot OAuth2 authentication backend"""
    name = 'hubspot'
    AUTHORIZATION_URL = 'https://app.hubspot.com/oauth/authorize'
    ACCESS_TOKEN_URL = 'https://api.hubapi.com/oauth/v1/token'
    ACCESS_TOKEN_METHOD = 'POST'
    USER_DATA_URL = 'https://api.hubapi.com/oauth/v1/access-tokens/'
    DEFAULT_SCOPE = ['oauth']
    EXTRA_DATA = [
        ('hub_domain', 'hub_domain'),
        ('hub_id', 'hub_id'),
        ('app_id', 'app_id'),
        ('user_id', 'user_id'),
        ('refresh_token', 'refresh_token'),
        ('expires_in', 'expires')
    ]

    def get_user_details(self, response):
        """Return user details"""
        response['email'] = response['user']
        return response

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data information from service"""
        return self.get_json(self.USER_DATA_URL + access_token, headers={
          'Authorization': 'Bearer ' + access_token
        })
